{ File   : Handler.pas
  Author : Deniz Oezmen
  Created: 2007-07-23
  Changed: 2007-07-26

  IMA ADPCM manager handler.
}
unit Handler;

interface

uses
  Manager;

const
  Handler_Success       = Longint($00000000);
  Handler_InvalidHandle = Longint($80010001);
  Handler_OutOfMemory   = Longint($80010002);

type
  THandler = record
    Assigned: Boolean;
    Manager : TIMA_ADPCM_Manager;
  end;

function GetNewHandle(BlockSize, SampPerBlock, NumChannels: Longint;
  const FormatType: TGUID): Longint;
function ReleaseHandle(Handle: Longint): Longint;
function GetHandler(Handle: Longint; out Handler: THandler): Longint;

implementation

uses
  SysUtils;

var
  Handlers: array of THandler;

function IsHandleValid(Handle: Longint): Longint;
begin
  Result := Handler_InvalidHandle;

  if (Handle < 0) or (Handle > High(Handlers)) then
    Exit;
  if not Handlers[Handle].Assigned then
    Exit;

  Result := Handler_Success
end;

function GetNewHandle(BlockSize, SampPerBlock, NumChannels: Longint;
  const FormatType: TGUID): Longint;
var
  i: Longint;
begin
  Result := -1;

  for i := 0 to High(Handlers) do
    if IsHandleValid(i) = Handler_InvalidHandle then
    begin
      Result := i;
      Break
    end;

  if Result = -1 then
  begin
    try
      SetLength(Handlers, Length(Handlers) + 1)
    except
      on EOutOfMemory do
      begin
        Result := Handler_OutOfMemory;
        Exit
      end
    end;
    Result := High(Handlers)
  end;

  Handlers[Result].Assigned := True;
  Handlers[Result].Manager := TIMA_ADPCM_Manager.Create(BlockSize,
    SampPerBlock, NumChannels, FormatType)
end;

function ReleaseHandle(Handle: Longint): Longint;
var
  i,
  LastAssigned: Longint;
begin
  Result := IsHandleValid(Handle);
  if Result < Handler_Success then
    Exit;

  Handlers[Handle].Assigned := False;
  Handlers[Handle].Manager.Free;
  Handlers[Handle].Manager := nil;

  if Handle = High(Handlers) then
    try
      SetLength(Handlers, Length(Handlers) - 1)
    except
      on EOutOfMemory do
      begin
        Result := Handler_OutOfMemory;
        Exit
      end
    end;

  LastAssigned := -1;
  for i := 0 to High(Handlers) do
    if Handlers[i].Assigned then
      LastAssigned := i;

  SetLength(Handlers, LastAssigned + 1);

  Result := Handler_Success
end;

procedure ReleaseAllHandles;
var
  i: Longint;
begin
  for i := 0 to High(Handlers) do
    if Handlers[i].Assigned then
      ReleaseHandle(i)
end;

function GetHandler(Handle: Longint; out Handler: THandler): Longint;
begin
  Result := IsHandleValid(Handle);
  if Result < Handler_Success then
    Exit;

  Handler := Handlers[Handle];

  Result := Handler_Success
end;

initialization
  SetLength(Handlers, 0);

finalization
  ReleaseAllHandles;

  Finalize(Handlers, 0);
  SetLength(Handlers, 0);
  Handlers := nil

end.

