{ File    : $URL: file:///D:/SVN/svnDelphi/tags/IMA/IMAMgr/1.10.5502.2/src/formats/Jack.pas $
  Author  : $Author: Deniz Özmen $
  Changed : $Date: 2011-07-02 23:57:59 +0200 (Sa, 02 Jul 2011) $
  Revision: $Rev: 23 $

  Jack Orlando-compatible IMA ADPCM routines.
}
unit Jack;

// {$DEFINE USE_INDEX_ADAPTATION}

interface

const
  PluginGUID: TGUID = '{6259E1BC-2212-4684-9422-3B982E10BC4C}';
  PluginDesc: string = 'Jack Orlando (compatible)';

implementation

uses
  Shared;

type
  TIMA_ADPCM_BlockHeader = packed record
    Sample  : Smallint;
    Index   : Word;
  end;

var
  PluginIndex    : Longint;
  PluginFunctions: TPluginFunctions;

function GetNumSamples(BlockSize, SampPerBlock, NumChannels: Longint): Longint;
begin
  Result := SampPerBlock
end;

function GetBlockSize(NumSamples: Longint; NumChannels: Longint): Longint;
begin
  if NumSamples >= NumChannels then
  begin
    Result := NumChannels * SizeOf(TIMA_ADPCM_BlockHeader) +
      (NumSamples div (8 * NumChannels) +
      Byte(NumSamples mod (8 * NumChannels) <> 0)) * 4 * NumChannels
  end
  else
    Result := 0
end;

procedure CorrectParams(var BlockSize, SampPerBlock, NumChannels: Longint);
begin
  if BlockSize > 0 then
    SampPerBlock := GetNumSamples(BlockSize, SampPerBlock, NumChannels)
  else
    BlockSize := GetBlockSize(SampPerBlock, NumChannels)
end;

function GetCompressedSize(BlockSize, SampPerBlock, NumChannels,
  NumSamples: Longint): Longint;
begin
  BlockSize := GetBlockSize(SampPerBlock, NumChannels);

  Result := (NumSamples div SampPerBlock +
    Byte(NumSamples mod SampPerBlock <> 0)) * BlockSize
end;

function GetDecompressedSize(BlockSize, SampPerBlock, NumChannels,
  CodeSize: Longint): Longint;
begin
  BlockSize := GetBlockSize(SampPerBlock, NumChannels);

  Result := (CodeSize div BlockSize +
    Byte(CodeSize mod BlockSize <> 0)) * SampPerBlock
end;

function CompressToBuffer(const SampBuffer: TIMA_ADPCM_SampleBuffer;
  var SampBufIndex: Longint; const CodeBuffer: TIMA_ADPCM_CodeBuffer;
  var CodeBufIndex: Longint; const States: TIMA_ADPCM_States;
  BlockSize, SampPerBlock, NumChannels: Longint): Longint;
var
  i, j,
  Channel,
  NumSamples: Longint;
  Codes     : array of Longword;
  Header    : TIMA_ADPCM_BlockHeader;
begin
  Result := 0;

  if not ((SampBufIndex <= High(SampBuffer)) and (SampBuffer <> nil) and
    (Length(SampBuffer) >= NumChannels)) then
    Exit;

  // calculate number of samples to compress
  SampPerBlock := GetNumSamples(BlockSize, SampPerBlock, NumChannels);
  if SampBufIndex + SampPerBlock - 1 <= High(SampBuffer) then
    NumSamples := SampPerBlock
  else
    NumSamples := High(SampBuffer) - SampBufIndex + 1;

  if GetBlockSize(NumSamples, NumChannels) >
    Length(CodeBuffer) - CodeBufIndex then
    Exit;

  // prepare block headers
  for i := 0 to NumChannels - 1 do
  begin
    Header.Sample := SampBuffer[SampBufIndex];

    {$IFDEF USE_INDEX_ADAPTATION}
    if SampBufIndex + NumChannels <= High(SampBuffer) then
      Header.Index := GetBestIndex(Header.Sample,
        SampBuffer[SampBufIndex + NumChannels])
    else
      Header.Index := 0;
    {$ELSE}
    Header.Index := States[i].Index;
    {$ENDIF}

    Move(Header, CodeBuffer[CodeBufIndex], SizeOf(Header));
    Inc(CodeBufIndex, SizeOf(Header));

    States[i].PrevSample := Header.Sample;
    States[i].Index := Header.Index
  end;

  // compress samples
  SetLength(Codes, NumChannels);
  for i := 0 to NumChannels - 1 do
    Codes[i] := 0;

  i := 0;
  Channel := 0;

  while i < NumSamples do
  begin
    // collect eight samples for each channel
    for j := 1 to 8 * NumChannels do
    begin
      Codes[Channel] := Codes[Channel] shr 4;

      if i < NumSamples then
      begin
        Codes[Channel] := Codes[Channel] or
          (EncodeSample(States[Channel], SampBuffer[SampBufIndex]) shl 28);
        Inc(SampBufIndex);

        Inc(i)
      end;

      Inc(Channel);
      if Channel = NumChannels then
        Channel := 0
    end;

    // emit codes to buffer
    for j := 0 to NumChannels - 1 do
    begin
      Move(Codes[j], CodeBuffer[CodeBufIndex], 4);
      Inc(CodeBufIndex, 4);
      Codes[j] := 0
    end
  end;

  Result := GetBlockSize(NumSamples, NumChannels)
end;

function DecompressFromBuffer(const SampBuffer: TIMA_ADPCM_SampleBuffer;
  var SampBufIndex: Longint; const CodeBuffer: TIMA_ADPCM_CodeBuffer;
  var CodeBufIndex: Longint; const States: TIMA_ADPCM_States;
  BlockSize, SampPerBlock, NumChannels: Longint): Longint;
var
  i, j,
  Channel,
  NumSamples: Longint;
  Codes     : array of Longword;
  Header    : TIMA_ADPCM_BlockHeader;
begin
  Result := 0;

  if not ((CodeBufIndex <= High(CodeBuffer)) and (CodeBuffer <> nil) and
    (Length(CodeBuffer) >= NumChannels * SizeOf(TIMA_ADPCM_BlockHeader))) then
    Exit;

  // calculate number of samples to decompress
  if CodeBufIndex + BlockSize - 1 <= High(CodeBuffer) then
    NumSamples := GetNumSamples(BlockSize, SampPerBlock, NumChannels)
  else
    NumSamples := GetNumSamples(High(CodeBuffer) - CodeBufIndex + 1,
      SampPerBlock, NumChannels);

  if NumSamples > Length(SampBuffer) - SampBufIndex then
    Exit;

  // read block headers
  for i := 0 to NumChannels - 1 do
  begin
    Move(CodeBuffer[CodeBufIndex], Header, SizeOf(Header));
    Inc(CodeBufIndex, SizeOf(Header));

    States[i].PrevSample := Header.Sample;
    States[i].Index := Header.Index
  end;

  // decompress samples
  SetLength(Codes, NumChannels);
  for i := 0 to NumChannels - 1 do
    Codes[i] := 0;

  i := 0;
  Channel := 0;

  while i < NumSamples do
  begin
    for j := 0 to NumChannels - 1 do
    begin
      Move(CodeBuffer[CodeBufIndex], Codes[j], 4);
      Inc(CodeBufIndex, 4)
    end;

    for j := 1 to 8 * NumChannels do
    begin
      if i < NumSamples then
      begin
        SampBuffer[SampBufIndex] := DecodeSample(States[Channel],
          Codes[Channel] and $0f);
        Inc(SampBufIndex);

        Inc(i)
      end;

      Codes[Channel] := Codes[Channel] shr 4;

      Inc(Channel);
      if Channel = NumChannels then
        Channel := 0
    end
  end;

  Result := NumSamples
end;

initialization
  PluginFunctions.CorrectParams := CorrectParams;
  PluginFunctions.GetCompressedSize := GetCompressedSize;
  PluginFunctions.GetDecompressedSize := GetDecompressedSize;
  PluginFunctions.CompressToBuffer := CompressToBuffer;
  PluginFunctions.DecompressFromBuffer := DecompressFromBuffer;

  PluginIndex := RegisterPlugin(PluginGUID, PluginDesc, PluginFunctions);

finalization
  UnregisterPlugin(PluginGUID);
  PluginIndex := -1
  
end.
