{ File    : $URL: file:///D:/SVN/svnDelphi/tags/IMA/IMAMgr/1.10.5502.2/src/formats/TLJ.pas $
  Author  : $Author: Deniz Özmen $
  Changed : $Date: 2011-07-02 23:57:59 +0200 (Sa, 02 Jul 2011) $
  Revision: $Rev: 23 $

  TLJ-compatible IMA ADPCM routines.
}
unit TLJ;

// {$DEFINE USE_INDEX_ADAPTATION}

interface

const
  PluginGUID: TGUID = '{24C8C90B-D945-4F5D-996A-15A9EDB46783}';
  PluginDesc: string = 'The Longest Journey (compatible)';

implementation

uses
  Shared;

type
  TIMA_ADPCM_BlockHeader = packed record
    Sample: Smallint;
    Index : Word;
  end;

var
  PluginIndex    : Longint;
  PluginFunctions: TPluginFunctions;

function GetNumSamples(BlockSize, SampPerBlock, NumChannels: Longint): Longint;
begin
  if BlockSize >= NumChannels * SizeOf(TIMA_ADPCM_BlockHeader) then
    Result := NumChannels + (BlockSize - NumChannels *
      SizeOf(TIMA_ADPCM_BlockHeader)) * 2
  else
    Result := 0
end;

function GetBlockSize(NumSamples: Longint; NumChannels: Longint): Longint;
begin
  if NumSamples >= NumChannels then
  begin
    Dec(NumSamples, NumChannels);
    Result := NumChannels * SizeOf(TIMA_ADPCM_BlockHeader) + NumSamples div 2 +
      NumSamples mod 2
  end
  else
    Result := 0
end;

procedure CorrectParams(var BlockSize, SampPerBlock, NumChannels: Longint);
begin
  if BlockSize > 0 then
    SampPerBlock := GetNumSamples(BlockSize, SampPerBlock, NumChannels)
  else
    BlockSize := GetBlockSize(SampPerBlock, NumChannels)
end;

function GetCompressedSize(BlockSize, SampPerBlock, NumChannels,
  NumSamples: Longint): Longint;
begin
  SampPerBlock := GetNumSamples(BlockSize, SampPerBlock, NumChannels);

  Result := (NumSamples div SampPerBlock) * BlockSize +
    GetBlockSize(NumSamples mod SampPerBlock, NumChannels)
end;

function GetDecompressedSize(BlockSize, SampPerBlock, NumChannels,
  CodeSize: Longint): Longint;
begin
  SampPerBlock := GetNumSamples(BlockSize, SampPerBlock, NumChannels);

  Result := (CodeSize div BlockSize) * SampPerBlock +
    GetNumSamples(CodeSize mod BlockSize, SampPerBlock, NumChannels)
end;

function CompressToBuffer(const SampBuffer: TIMA_ADPCM_SampleBuffer;
  var SampBufIndex: Longint; const CodeBuffer: TIMA_ADPCM_CodeBuffer;
  var CodeBufIndex: Longint; const States: TIMA_ADPCM_States;
  BlockSize, SampPerBlock, NumChannels: Longint): Longint;
var
  i,
  Channel,
  NumSamples: Longint;
  Code      : Byte;
  Header    : TIMA_ADPCM_BlockHeader;
begin
  Result := 0;

  if not ((SampBufIndex <= High(SampBuffer)) and (SampBuffer <> nil) and
    (Length(SampBuffer) >= NumChannels)) then
    Exit;

  // calculate number of samples to compress
  SampPerBlock := GetNumSamples(BlockSize, SampPerBlock, NumChannels);
  if SampBufIndex + SampPerBlock - 1 <= High(SampBuffer) then
    NumSamples := SampPerBlock
  else
    NumSamples := High(SampBuffer) - SampBufIndex + 1;

  if GetBlockSize(NumSamples, NumChannels) >
    Length(CodeBuffer) - CodeBufIndex then
    Exit;

  // prepare block headers
  for i := 0 to NumChannels - 1 do
  begin
    Header.Sample := SampBuffer[SampBufIndex];

    {$IFDEF USE_INDEX_ADAPTATION}
    if SampBufIndex + NumChannels <= High(SampBuffer) then
      Header.Index := GetBestIndex(Header.Sample,
        SampBuffer[SampBufIndex + NumChannels])
    else
      Header.Index := 0;
    {$ELSE}
    Header.Index := States[i].Index;
    {$ENDIF}

    Inc(SampBufIndex);

    Move(Header, CodeBuffer[CodeBufIndex], SizeOf(Header));
    Inc(CodeBufIndex, SizeOf(Header));

    States[i].PrevSample := Header.Sample;
    States[i].Index := Header.Index
  end;

  // compress samples
  i := 0;
  Channel := NumChannels - 1;
  Code := 0;
  Dec(NumSamples, NumChannels);
  while i < NumSamples do
  begin
    // lower nibble: high channel
    if (i and $00000001) = 0 then
      Code := EncodeSample(States[Channel], SampBuffer[SampBufIndex + Channel])
    else
      Code := Code or (EncodeSample(States[Channel],
        SampBuffer[SampBufIndex + Channel])) shl 4;

    // proceed to next channel
    if Channel > 0 then
      Dec(Channel)
    else
    begin
      Channel := NumChannels - 1;
      Inc(SampBufIndex, NumChannels)
    end;

    Inc(i);

    // emit to buffer if two 4-bit code words have been collected or end of
    // stream has been reached
    if (i and $00000001) = 0 then
    begin
      CodeBuffer[CodeBufIndex] := Code;
      Inc(CodeBufIndex)
    end
    else if i = NumSamples then
    begin
      CodeBuffer[CodeBufIndex] := Code shl 4;
      Inc(CodeBufIndex)
    end
  end;

  Result := GetBlockSize(NumSamples + NumChannels, NumChannels)
end;

function DecompressFromBuffer(const SampBuffer: TIMA_ADPCM_SampleBuffer;
  var SampBufIndex: Longint; const CodeBuffer: TIMA_ADPCM_CodeBuffer;
  var CodeBufIndex: Longint; const States: TIMA_ADPCM_States;
  BlockSize, SampPerBlock, NumChannels: Longint): Longint;
var
  i,
  Channel,
  NumSamples: Longint;
  Header    : TIMA_ADPCM_BlockHeader;
begin
  Result := 0;

  if not ((CodeBufIndex <= High(CodeBuffer)) and (CodeBuffer <> nil) and
    (Length(CodeBuffer) >= NumChannels * SizeOf(TIMA_ADPCM_BlockHeader))) then
    Exit;

  // calculate number of samples to decompress
  if CodeBufIndex + BlockSize - 1 <= High(CodeBuffer) then
    NumSamples := GetNumSamples(BlockSize, SampPerBlock, NumChannels)
  else
    NumSamples := GetNumSamples(High(CodeBuffer) - CodeBufIndex + 1,
      SampPerBlock, NumChannels);

  if NumSamples > Length(SampBuffer) - SampBufIndex then
    Exit;

  // read block headers
  for i := 0 to NumChannels - 1 do
  begin
    Move(CodeBuffer[CodeBufIndex], Header, SizeOf(Header));
    Inc(CodeBufIndex, SizeOf(Header));

    SampBuffer[SampBufIndex] := Header.Sample;
    Inc(SampBufIndex);

    States[i].PrevSample := Header.Sample;
    States[i].Index := Header.Index
  end;

  // decompress samples
  i := 0;
  Channel := NumChannels - 1;
  Dec(NumSamples, NumChannels);
  while i < NumSamples do
  begin
    // lower nibble: high channel
    if (i and $00000001) = 0 then
      SampBuffer[SampBufIndex + Channel] :=
        DecodeSample(States[Channel], CodeBuffer[CodeBufIndex] and $0f)
    else
      SampBuffer[SampBufIndex + Channel] :=
        DecodeSample(States[Channel], CodeBuffer[CodeBufIndex] shr 4);

    // proceed to next channel
    if Channel > 0 then
      Dec(Channel)
    else
    begin
      Channel := NumChannels - 1;
      Inc(SampBufIndex, NumChannels)
    end;

    if (i and $00000001) = 1 then
      Inc(CodeBufIndex);

    Inc(i)
  end;

  Result := NumSamples + NumChannels
end;

initialization
  PluginFunctions.CorrectParams := CorrectParams;
  PluginFunctions.GetCompressedSize := GetCompressedSize;
  PluginFunctions.GetDecompressedSize := GetDecompressedSize;
  PluginFunctions.CompressToBuffer := CompressToBuffer;
  PluginFunctions.DecompressFromBuffer := DecompressFromBuffer;

  PluginIndex := RegisterPlugin(PluginGUID, PluginDesc, PluginFunctions);

finalization
  UnregisterPlugin(PluginGUID);
  PluginIndex := -1

end.
